import json, yaml
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
import matplotlib.ticker as mtick
import numpy as np


def LoadJson(file_name):
    JSONPATH = os.path.join(file_name)
    return yaml.safe_load(open(JSONPATH))


def FormatFig(xlabel,ylabel,ax0,xpos=0.9,ypos=0.9):
    #Limit number of digits in ticks
    # y_loc, _ = plt.yticks()
    # y_update = ['%.1f' % y for y in y_loc]
    # plt.yticks(y_loc, y_update) 
    ax0.set_xlabel(xlabel,fontsize=24)
    ax0.set_ylabel(ylabel)

def SetStyle():
    from matplotlib import rc
    rc('text', usetex=True)

    import matplotlib as mpl
    rc('font', family='serif')
    rc('font', size=22)
    rc('xtick', labelsize=15)
    rc('ytick', labelsize=14)
    rc('legend', fontsize=15)

    # #
    mpl.rcParams.update({'font.size': 19})
    #mpl.rcParams.update({'legend.fontsize': 18})
    mpl.rcParams['text.usetex'] = False
    mpl.rcParams.update({'xtick.labelsize': 18}) 
    mpl.rcParams.update({'ytick.labelsize': 18}) 
    mpl.rcParams.update({'axes.labelsize': 18}) 
    mpl.rcParams.update({'legend.frameon': False}) 
    
    import matplotlib.pyplot as plt

def SetGrid(ratio=True):
    fig = plt.figure(figsize=(9, 9))
    if ratio:
        gs = gridspec.GridSpec(2, 1, height_ratios=[3,1]) 
        gs.update(wspace=0.025, hspace=0.1)
    else:
        gs = gridspec.GridSpec(1, 1)
    return fig,gs


def HistRoutine(feed_dict,
                xlabel='',
                ylabel='Normalized number of events',
                reference_name='data',
                logy=False,
                logx = False,
                binning=None,
                label_loc='best',
                plot_ratio=True,
                weights=None,
                #colors from https://github.com/mpetroff/accessible-color-cycles
                color_list = ['black',"#3f90da", "#ffa90e", "#bd1f01",
                              "#94a4a2", "#832db6", "#a96b59", "#e76300",
                              "#b9ac70", "#717581", "#92dadd"],
                ref_plot = {'histtype':'stepfilled','alpha':0.2},
                other_plots = {'histtype':'step','linewidth':2},
                marker_style = 'o',
                uncertainty=None):
    if plot_ratio:
        assert reference_name in feed_dict.keys(), "ERROR: Don't know the reference distribution"

    fig,gs = SetGrid(ratio=plot_ratio) 
    ax0 = plt.subplot(gs[0])

    color = {}
    if plot_ratio:
        plt.xticks(fontsize=0)
        ax1 = plt.subplot(gs[1],sharex=ax0)

    
    if binning is None:
        binning = np.linspace(np.quantile(feed_dict[reference_name],0.01),np.quantile(feed_dict[reference_name],0.99),50)
        
    xaxis = [(binning[i] + binning[i+1])/2.0 for i in range(len(binning)-1)]

    if weights is not None:
        reference_hist,_ = np.histogram(feed_dict[reference_name],weights=weights[reference_name],bins=binning,density=True)
    else:
        reference_hist,_ = np.histogram(feed_dict[reference_name],bins=binning,density=True)

    maxy = 0    
    for ip,plot in enumerate(feed_dict.keys()):
        color[plot] = color_list[ip]
        plot_style = ref_plot if reference_name == plot else other_plots
        if weights is not None:
            dist,_,_=ax0.hist(feed_dict[plot],bins=binning,label=plot,density=True,color=color[plot],
                              weights=weights[plot],**plot_style)
        else:
            dist,_,_=ax0.hist(feed_dict[plot],bins=binning,label=plot,density=True,color=color[plot],**plot_style)

        if np.max(dist) > maxy:
            maxy = np.max(dist)
            
        if plot_ratio:
            if reference_name!=plot:
                ratio = np.ma.divide(dist,reference_hist).filled(0)
                ax1.plot(xaxis,ratio,ms=10,marker = marker_style,color=color[plot],
                         lw=0,markerfacecolor='none',markeredgewidth=3)
                if uncertainty is not None:
                    for ibin in range(len(binning)-1):
                        xup = binning[ibin+1]
                        xlow = binning[ibin]
                        ax1.fill_between(np.array([xlow,xup]),
                                         uncertainty[ibin],-uncertainty[ibin], alpha=0.3,color='k')    
    if logy:
        ax0.set_yscale('log')
        ax0.set_ylim(1e-5,10*maxy)
    else:
        ax0.set_ylim(0,1.3*maxy)

    if logx:
        #ax0.set_xscale('log')
        ax1.set_xscale('log')

    ax0.legend(loc=label_loc,fontsize=16,ncol=2)
    if plot_ratio:
        FormatFig(xlabel = "", ylabel = ylabel,ax0=ax0) 
        plt.ylabel('Ratio to Data')
        plt.axhline(y=1.0, color='r', linestyle='-',linewidth=1)
        plt.ylim([0.5,1.5])
        plt.xlabel(xlabel)
    else:
        FormatFig(xlabel = xlabel, ylabel = ylabel,ax0=ax0) 
        
    return fig,ax0



from  omnifold import DataLoader
import numpy as np

mock_dataset = np.zeros((100,3,4))
mock_dataloader = DataLoader(reco = mock_dataset,gen = mock_dataset,normalize=True)
pass_reco = [0 if i % 2 == 0 else 1 for i in range(mock_dataloader.reco.shape[0])]

mock_dataloader = DataLoader(reco = mock_dataset,pass_reco = pass_reco,
                             normalize=True)
print(mock_dataloader.weight)



import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.callbacks import EarlyStopping,ModelCheckpoint, ReduceLROnPlateau
import sys, os
import horovod.tensorflow.keras as hvd

from datetime import datetime
import gc
import pickle
from omnifold.net import weighted_binary_crossentropy


def expit(x):
    return 1. / (1. + np.exp(-x))

class MultiFold():
    def __init__(self,
                 name,
                 model_reco,
                 model_gen,
                 data,
                 mc,
                 weights_folder = 'weights',
                 nstrap=0,
                 niter = 3,
                 batch_size = 128,
                 epochs = 50,
                 lr = 1e-4,
                 early_stop = 10,
                 start = 0,
                 train_frac = 0.8,
                 size = 1,
                 rank = 0,
                 verbose=False):

        self.name=name
        self.niter = niter
        self.data = data
        self.mc = mc
        self.nstrap = nstrap
        self.start = start
        self.train_frac = train_frac
        self.size = size
        self.rank = rank
        self.verbose = verbose*(self.rank==0)
        self.log_file =  open('log_{}.txt'.format(self.name),'w')
        
        #Model specific parameters
        self.model1 = model_reco
        self.model2 = model_gen
        self.BATCH_SIZE=batch_size
        self.EPOCHS=epochs
        self.LR = lr
        self.patience = early_stop

        self.num_steps_reco = None
        self.num_steps_gen = None
                
        self.weights_folder = weights_folder
        if self.nstrap>0:
            self.weights_folder = f'{self.weights_folder}_strap'
            
        if not os.path.exists(self.weights_folder):
            os.makedirs(self.weights_folder)

        if not hvd.is_initialized():
            hvd.init()

    def Unfold(self):                                        
        self.weights_pull = np.ones(self.mc.weight.shape[0],dtype=np.float32)
        if self.start>0:
            if self.verbose: self.log_string(f"INFO: Continuing OmniFold training from Iteration {self.start}")
            if self.rank == 0:
                self.log_string("Loading step 2 weights from iteration {}".format(self.start-1))
            model_name = '{}/OmniFold_{}_iter{}_step2/checkpoint'.format(self.weights_folder,self.name,self.start-1)
            self.model2.load_weights(model_name).expect_partial()
            self.weights_push = self.reweight(self.mc.gen,self.model2,batch_size=1000)
            #Also load model 1 to have a better starting point
            model_name = '{}/OmniFold_{}_iter{}_step1/checkpoint'.format(self.weights_folder,self.name,self.start-1)
            self.model1.load_weights(model_name).expect_partial()
        else:
            self.weights_push = np.ones(self.mc.weight.shape[0],dtype=np.float32)

        self.CompileModel()
        for i in range(self.start,self.niter):
            if self.rank==0:
                self.log_string("ITERATION: {}".format(i + 1))
            self.RunStep1(i)        
            self.RunStep2(i)
            self.CompileModel(fixed=True)

    def RunStep1(self,i):
        '''Data versus reco MC reweighting'''
        if self.rank==0:
            self.log_string("RUNNING STEP 1")
        
        self.RunModel(
            np.concatenate((self.labels_mc,self.labels_data)),
            
            np.concatenate((self.weights_push*self.mc.weight*self.mc.pass_reco,
                            self.data.weight*self.data.pass_reco)),
            
            i,self.model1,stepn=1,
            NTRAIN = self.num_steps_reco*self.BATCH_SIZE,
            cached = i>self.start #after first training cache the training data
        )

        #Don't update weights where there's no reco events
        new_weights = np.ones_like(self.weights_pull)
        new_weights[self.mc.pass_reco] = self.reweight(self.mc.reco,self.model1,batch_size=1000)[self.mc.pass_reco]
        self.weights_pull = self.weights_push *new_weights

    def RunStep2(self,i):
        '''Gen to Gen reweighing'''        
        if self.rank==0:
            self.log_string("RUNNING STEP 2")
        
        self.RunModel(
            np.concatenate((self.labels_mc, self.labels_gen)),
            np.concatenate((self.mc.weight*self.mc.pass_gen, self.mc.weight*self.weights_pull*self.mc.pass_gen)),
            i,self.model2,stepn=2,
            NTRAIN = self.num_steps_gen*self.BATCH_SIZE,
            cached = i>self.start #after first training cache the training data
        )
        new_weights = np.ones_like(self.weights_push)
        new_weights[self.mc.pass_gen]=self.reweight(self.mc.gen,self.model2)[self.mc.pass_gen]
        self.weights_push = new_weights



    def RunModel(self,
                 labels,
                 weights,
                 iteration,
                 model,
                 stepn,
                 NTRAIN=1000,
                 cached = False,
                 ):

        test_frac = 1.-self.train_frac
        NTEST = int(test_frac*NTRAIN)
        train_data, test_data = self.cache(labels,weights,stepn,cached,NTRAIN-NTEST)
        
        if self.verbose:
            print(80*'#')
            self.log_string("Train events used: {}, Test events used: {}".format(NTRAIN,NTEST))
            print(80*'#')


        
        callbacks = [
            hvd.callbacks.BroadcastGlobalVariablesCallback(0),
            hvd.callbacks.MetricAverageCallback(),

            #ReduceLR only used to keep track of the LR during training
            ReduceLROnPlateau(patience=1000, min_lr=1e-7,
                              verbose=self.verbose,
                              monitor="val_loss"),
            EarlyStopping(patience=self.patience,
                          restore_best_weights=True,
                          monitor="val_loss"),
        ]
        
        
        if self.rank ==0:
            if self.nstrap>0:
                model_name = '{}/OmniFold_{}_iter{}_step{}_strap{}/checkpoint'.format(
                    self.weights_folder,self.name,iteration,stepn,self.nstrap)
            else:
                model_name = '{}/OmniFold_{}_iter{}_step{}/checkpoint'.format(
                    self.weights_folder,self.name,iteration,stepn)
                
            callbacks.append(ModelCheckpoint(model_name,
                                             save_best_only=True,
                                             mode='auto',period=1,
                                             save_weights_only=True))
                    
        hist =  model.fit(
            train_data,
            epochs=self.EPOCHS,
            steps_per_epoch=int(self.train_frac*NTRAIN//self.BATCH_SIZE),
            validation_data= test_data,
            validation_steps=NTEST//self.BATCH_SIZE,
            verbose= self.verbose,
            callbacks=callbacks)
        
        self.log_string(f"Last val loss {hist.history['val_loss'][0]}")
        
        if self.rank ==0:
            self.log_string("INFO: Dumping training history ...")
            with open(model_name.replace("/checkpoint",".pkl"),"wb") as f:
                pickle.dump(hist.history, f)
        
        del train_data, test_data
        gc.collect()


    def cache(self,
              label,
              weights,
              stepn,
              cached,
              NTRAIN
              ):


        if not cached:
            if self.verbose:
                self.log_string("Creating cached data from step {}".format(stepn))
                    
            if stepn ==1:
                self.idx_1 = np.arange(label.shape[0])
                np.random.shuffle(self.idx_1)
                self.tf_data1 = tf.data.Dataset.from_tensor_slices(
                    np.concatenate([self.mc.reco,self.data.reco],0)[self.idx_1])
            elif stepn ==2:
                self.idx_2 = np.arange(label.shape[0])
                np.random.shuffle(self.idx_2)
                self.tf_data2 = tf.data.Dataset.from_tensor_slices(
                    np.concatenate([self.mc.gen,self.mc.gen],0)[self.idx_2])
                
        idx = self.idx_1 if stepn==1 else self.idx_2
        labels = tf.data.Dataset.from_tensor_slices(np.stack((label[idx],weights[idx]),axis=1))
        
        if stepn ==1:
            data = tf.data.Dataset.zip((self.tf_data1,labels))
        elif stepn==2:
            data = tf.data.Dataset.zip((self.tf_data2,labels))
        else:
            logging.error("ERROR: STEPN not recognized")

                
        train_data = data.take(NTRAIN).shuffle(NTRAIN).repeat().batch(self.BATCH_SIZE).prefetch(buffer_size=tf.data.AUTOTUNE)
        test_data  = data.skip(NTRAIN).repeat().batch(self.BATCH_SIZE).prefetch(buffer_size=tf.data.AUTOTUNE)
        del data
        gc.collect()
        return train_data, test_data

    def Preprocessing(self):
        self.PrepareInputs()

    def get_optimizer(self,num_steps,fixed=False,min_learning_rate = 1e-5):
        opt = tf.keras.optimizers.Adam(learning_rate=min_learning_rate if fixed else self.LR)
        opt = hvd.DistributedOptimizer(opt)
        return opt
        

    def CompileModel(self,fixed=False):

        if self.num_steps_reco ==None:
            self.num_steps_reco = int((self.mc.nmax + self.data.nmax))//self.size//self.BATCH_SIZE
            self.num_steps_gen = 2*self.mc.nmax//self.size//self.BATCH_SIZE
            if self.rank==0:
                self.log_string(f"{self.num_steps_reco} training steps at reco and {self.num_steps_gen} steps at gen")


        opt1 = self.get_optimizer(int(self.train_frac*self.num_steps_reco),fixed=fixed)
        opt2 = self.get_optimizer(int(self.train_frac*self.num_steps_gen),fixed=fixed)
        

        self.model1.compile(opt1,experimental_run_tf_function=False,
                            loss = weighted_binary_crossentropy,
                            weighted_metrics=[])
        self.model2.compile(opt2,experimental_run_tf_function=False,
                            loss = weighted_binary_crossentropy,
                            weighted_metrics=[])


    def PrepareInputs(self):
        self.labels_mc = np.zeros(len(self.mc.pass_reco),dtype=np.float32)
        self.labels_data = np.ones(len(self.data.pass_reco),dtype=np.float32)
        self.labels_gen = np.ones(len(self.mc.pass_gen),dtype=np.float32)

    def reweight(self,events,model,batch_size=None):
        if batch_size is None:
           batch_size =  self.BATCH_SIZE

        f = expit(model.predict(events,batch_size=batch_size,verbose=self.verbose))
        weights = f / (1. - f)
        return np.nan_to_num(weights[:,0],posinf=1)

        
    def log_string(self,out_str):
        self.log_file.write(out_str+'\n')
        self.log_file.flush()
        print(out_str)






from tensorflow.keras.layers import Dense, Input, Dropout
import tensorflow.keras.layers as layers
from tensorflow.keras.models import Model
import tensorflow as tf
from tensorflow import keras


def weighted_binary_crossentropy(y_true, y_pred):
    weights = tf.gather(y_true, [1], axis=1) # event weights
    y_true = tf.gather(y_true, [0], axis=1) # actual y_true for loss

    t_loss = weights*tf.nn.sigmoid_cross_entropy_with_logits(labels=y_true, logits=y_pred)
    return tf.reduce_mean(t_loss)


def MLP(nvars,
        layer_sizes = [64,128,64],
        activation = 'gelu'):    
    '''
    Define a simple fully conneted model to be used during unfolding
    Function Inputs:
    nvars (int): The number of input variables for the model, defining the dimensionality of the input data.
    layer_sizes (list of int, optional, default=[64, 128, 64]): A list specifying the number of neurons in each hidden layer. The length of the list determines the number of layers, and each value corresponds to the number of units in that layer.
    activation (str, optional, default='gelu'): The activation function applied to each hidden layer. By default, it is set to 'gelu' (Gaussian Error Linear Unit), which adds non-linearity to the model.
    '''

    inputs = Input((nvars, ))
    layer = Dense(layer_sizes[0],activation=activation)(inputs)
    for layer_size in layer_sizes[1:]:
        layer = Dense(layer_size,activation=activation)(layer)
        
    outputs = Dense(1)(layer)
    model = Model(inputs = inputs, outputs = outputs)
    return model

class PET(Model):
    """OmniFold Classifier class"""
    def __init__(self,
                 num_feat,
                 num_evt = 0,
                 num_part=100,
                 num_heads=2,
                 num_transformer= 2,
                 projection_dim= 32,
                 local = True,
                 K = 3,
                 layer_scale = True,
                 layer_scale_init = 1e-3
                 ):
        super(PET, self).__init__()
        '''
        Class Variables:
        num_feat (int): The number of input features per particle.
        num_evt (int, optional, default=0): Number of high-level featuresto use. If 0, no high level feature is considered
        num_part (int, optional, default=100): Number of particles or objects per event in the input data.
        num_heads (int, optional, default=2): Number of attention heads in the transformer layer, for multi-head self-attention.
        num_transformer (int, optional, default=2): Number of transformer layers stacked in the model.
        projection_dim (int, optional, default=32): Dimensionality of the projection layer for transforming input features.
        local (bool, optional, default=True): Flag to indicate whether a local GNN using k-nearest neighbors is used..
        K (int, optional, default=3): Parameter controlling the number of k-nearest neighbors to use.
        layer_scale (bool, optional, default=True): Determines whether layer scaling is applied for stabilizing training.
        layer_scale_init (float, optional, default=1e-3): Initial value for the layer scaling factor to ensure stable gradient behavior.
        '''
        
        self.num_feat = num_feat
        self.num_evt = num_evt
        self.num_part = num_part
        inputs_part = layers.Input((self.num_part,self.num_feat))
        if local:
            assert self.num_part >= K, "ERROR: K neighbors exceeding the number of particles"
            
        if self.num_evt>0:
            inputs_evt = layers.Input((self.num_evt))
        else:
            inputs_evt =  None
                        
        outputs_body = self.PET_body(inputs_part,
                                     num_part,
                                     num_heads,
                                     num_transformer,
                                     projection_dim,
                                     local = local, K = K,
                                     layer_scale = layer_scale,
                                     layer_scale_init = layer_scale_init
                                     )
                
        outputs_head = self.PET_head(outputs_body,inputs_evt,projection_dim)

        if inputs_evt is None:
            self.model = Model(inputs=inputs_part,
                               outputs=outputs_head)
        else:
            self.model = Model(inputs=inputs_part,
                               outputs=outputs_head)

        self.loss_tracker = keras.metrics.Mean(name="loss")

    @property
    def metrics(self):
        """List of the model's metrics.
        We make sure the loss tracker is listed as part of `model.metrics`
        so that `fit()` and `evaluate()` are able to `reset()` the loss tracker
        at the start of each epoch and at the start of an `evaluate()` call.
        """
        return [self.loss_tracker]

    def call(self,x, training=True):
        return self.model(x,training)
        

    def PET_body(
            self,
            inputs_part,
            num_parts,
            num_heads=4,
            num_transformer= 8,
            projection_dim= 64,
            local = True, K = 5,
            num_local = 2,
            layer_scale = True,
            layer_scale_init = 1e-3,
    ):
    
        encoded = get_encodding(inputs_part,projection_dim)
        inputs_mask = tf.cast(inputs_part[:,:,0,None] != 0, dtype='float32')
        if local:
            coord_shift = tf.multiply(999., tf.cast(tf.equal(inputs_mask, 0), dtype='float32'))        
            points = inputs_part[:,:,:2] #assume first 2 coordinates are eta-phi
            local_features = inputs_part
            for _ in range(num_local):
                local_features = get_neighbors(coord_shift+points,local_features,projection_dim,K)
                points = local_features
                
            encoded = layers.Add()([local_features,encoded])
            
        skip_connection = encoded*inputs_mask                     
        for i in range(num_transformer):
            x1 = layers.GroupNormalization(groups=1)(encoded)
            updates = layers.MultiHeadAttention(num_heads=num_heads,
                                                key_dim=projection_dim//num_heads)(x1,x1)
                
            if layer_scale:
                updates = LayerScale(layer_scale_init, projection_dim)(updates,inputs_mask)
            
            x2 = layers.Add()([updates,encoded])
            x3 = layers.GroupNormalization(groups=1)(x2)
            x3 = layers.Dense(2*projection_dim,activation="gelu")(x3)
            x3 = layers.Dense(projection_dim)(x3)
            if layer_scale:
                x3 = LayerScale(layer_scale_init, projection_dim)(x3,inputs_mask)
            encoded = layers.Add()([x3,x2])*inputs_mask
                        
        return encoded + skip_connection


    def PET_head(
            self,
            encoded,
            input_evt,
            projection_dim= 64,
            num_heads=4,
            num_class_layers=2,
            layer_scale = True,
            layer_scale_init = 1e-3,
    ):


        if input_evt is not None:
            conditional = layers.Dense(2*projection_dim,activation='gelu')(input_evt)
            conditional = tf.tile(conditional[:,None, :], [1,tf.shape(encoded)[1], 1])
            scale,shift = tf.split(conditional,2,-1)
            encoded = encoded*(1.0 + scale) + shift

        class_tokens = tf.Variable(tf.zeros(shape=(1, projection_dim)),trainable = True)    
        class_tokens = tf.tile(class_tokens[None, :, :], [tf.shape(encoded)[0], 1, 1])
        
        for _ in range(num_class_layers):
            concatenated = tf.concat([class_tokens, encoded],1)

            x1 = layers.GroupNormalization(groups=1)(concatenated)            
            updates = layers.MultiHeadAttention(num_heads=num_heads,
                                                key_dim=projection_dim//num_heads)(
                                                    query=x1[:,:1], value=x1, key=x1)
            updates = layers.GroupNormalization(groups=1)(updates)
            if layer_scale:
                updates = LayerScale(layer_scale_init, projection_dim)(updates)

            x2 = layers.Add()([updates,class_tokens])
            x3 = layers.GroupNormalization(groups=1)(x2)
            x3 = layers.Dense(2*projection_dim,activation="gelu")(x3)
            x3 = layers.Dense(projection_dim)(x3)
            if layer_scale:
                x3 = LayerScale(layer_scale_init, projection_dim)(x3)
            class_tokens = layers.Add()([x3,x2])

        class_tokens = layers.GroupNormalization(groups=1)(class_tokens)
        outputs_pred = layers.Dense(1,activation=None)(class_tokens[:,0])
                
        return outputs_pred


     def train_step(self, inputs):
        x,y = inputs
        with tf.GradientTape(persistent=True) as tape:
            y_pred = self.mdoel(x)
            loss = weighted_binary_crossentropy(y, y_pred)

        self.optimizer.minimize(loss,self.model.trainable_variables,tape=tape)        
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}

    def test_step(self, inputs):
        x,y = inputs            
        y_pred = self.model(x)
        loss = weighted_binary_crossentropy(y, y_pred)
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}
        

def get_neighbors(points,features,projection_dim,K):
    drij = pairwise_distance(points)  # (N, P, P)
    _, indices = tf.nn.top_k(-drij, k=K + 1)  # (N, P, K+1)
    indices = indices[:, :, 1:]  # (N, P, K)
    knn_fts = knn(tf.shape(points)[1], K, indices, features)  # (N, P, K, C)
    knn_fts_center = tf.broadcast_to(tf.expand_dims(features, 2), tf.shape(knn_fts))
    local = tf.concat([knn_fts-knn_fts_center,knn_fts_center],-1)
    local = layers.Dense(2*projection_dim,activation='gelu')(local)
    local = layers.Dense(projection_dim,activation='gelu')(local)
    local = tf.reduce_mean(local,-2)
    
    return local


def pairwise_distance(point_cloud):
    r = tf.reduce_sum(point_cloud * point_cloud, axis=2, keepdims=True)
    m = tf.matmul(point_cloud, point_cloud, transpose_b = True)
    D = r - 2 * m + tf.transpose(r, perm=(0, 2, 1))
    return tf.abs(D)


def knn(num_points, k, topk_indices, features):
    # topk_indices: (N, P, K)
    # features: (N, P, C)    
    batch_size = tf.shape(features)[0]

    batch_indices = tf.reshape(tf.range(batch_size), (-1, 1, 1))
    batch_indices = tf.tile(batch_indices, (1, num_points, k))
    indices = tf.stack([batch_indices, topk_indices], axis=-1)
    return tf.gather_nd(features, indices)


def get_encodding(x,projection_dim):
    x = layers.Dense(2*projection_dim,activation='gelu')(x)
    x = layers.Dense(projection_dim,activation='gelu')(x)
    return x

class LayerScale(layers.Layer):
    def __init__(self, init_values, projection_dim, **kwargs):
        super(LayerScale, self).__init__(**kwargs)
        self.init_values = init_values
        self.projection_dim = projection_dim
        self.gamma_initializer = tf.keras.initializers.Constant(self.init_values)

    def build(self, input_shape):
        # Ensure the layer is properly built by defining its weights in the build method
        self.gamma = self.add_weight(
            shape=(self.projection_dim,),
            initializer=self.gamma_initializer,
            trainable=True,
            name='gamma'
        )
        super(LayerScale, self).build(input_shape)


    def call(self, inputs,mask=None):
        # Element-wise multiplication of inputs and gamma
        if mask is not None:
            return inputs * self.gamma* mask
        else:
            return inputs * self.gamma

    def get_config(self):
        # Returns the configuration of the layer for serialization
        config = super(LayerScale, self).get_config()
        config.update({
            'init_values': self.init_values,
            'projection_dim': self.projection_dim
        })
        return config




import numpy as np
import sys, os



class DataLoader():
    def __init__(
            self,
            reco,            
            pass_reco = None,
            gen = None,
            pass_gen = None,
            weight = None,
            normalize=False,
            normalization_factor = 1_000_000,
    ):
        self.reco = reco
        self.pass_reco = pass_reco
        self.gen = gen
        self.pass_gen = pass_gen

        self.nmax = self.reco.shape[0]
        self.weight = weight
        
        if self.weight is None:
            print("INFO: Creating weights ...")
            self.weight = np.ones(reco.shape[0],dtype=np.float32)
            
        if self.pass_reco is None:
            print("INFO: Creating pass reco flag ...")
            self.pass_reco = np.ones(reco.shape[0],dtype=bool)
        else:
            #Make a boolean mask
            self.pass_reco = np.array(self.pass_reco) == 1
            
        if self.gen is not None:
            self.is_mc = True
        else:
            self.is_mc = False
            
        if self.is_mc and self.pass_gen is None:
            print("INFO: Creating pass gen flag ...")
            self.pass_gen = np.ones(gen.shape[0],dtype=bool)
        else:
            #Make a boolean mask
            self.pass_gen = np.array(self.pass_gen) == 1


        if normalize:
            print(f"INFO: Normalizing sum of weights to {normalization_factor} ...")
            sumw = np.sum(self.weight[np.array(self.pass_reco)==1])
            self.weight *= (normalization_factor/sumw).astype(np.float32)



# setup.py
from setuptools import setup, find_packages

setup(
    name='omnifold',
    version='0.1.9',
    packages=find_packages(),
    install_requires=[
        'horovod==0.28.1',
        'matplotlib',
        'numpy',
        'PyYAML',
        'setuptools',
        'tensorflow',
    ],
    author='Vinicius Mikuni',
    author_email='vmikuni@lbl.gov',
    description='OmniFold, a library to perform unbinned and high-dimensional unfolding for HEP.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/ViniciusMikuni/omnifold',  # Update with your GitHub URL
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
